from pymbse.commons import json_file


def get_design_variable_from_local_variable_file_cache(design_variables_mode,
                                                       design_variables,
                                                       url="http://cache_rest:8000",
                                                       get_max_optimization_ix=None,
                                                       get_best_optimization_snapshot=None):
    if design_variables_mode == "LOCAL_FILE":
        return json_file.read("../input/design_variables.json")
    elif design_variables_mode == "BEST_SNAPSHOT":
        max_opt_ix = get_max_optimization_ix(url)
        return get_best_optimization_snapshot(url, max_opt_ix)["input_parameters"]["design_variables"]
    elif design_variables_mode == "VARIABLE":
        return design_variables
    else:
        raise AttributeError(f"Design variables mode {design_variables_mode} not supported")



import os
import datetime
from abc import abstractmethod, ABC
from typing import Union

from pydantic import BaseModel

from pymbse.query.config.notebook_script_config import ScriptConfig
from pymbse.query.model_api.snapshot.model_snapshot import ModelSnapshot
from pymbse.commons import text_file
from pymbse.query.config.pymbse_config import NotebookScriptConfig


FIGURES_OF_MERIT_KEYWORD = "fom"
ARTEFACTS_KEYWORD = "artefacts"


class Executor(BaseModel, ABC):
    """Abstract dataclass storing a model dependency config container and parameters dictionary needed for model
    execution. The execute method is abstract and implemented per each subclass.

    """

    model_dependency_config: Union[ScriptConfig, NotebookScriptConfig]
    input_parameters: dict
    input_files: dict

    @abstractmethod
    def execute(self) -> dict:
        """Method executing a model (e.g., a notebook, a scripted notebook) and returning a dictionary with figures of
        merit and artifact files

        :return: a dictionary with figures of merit and artefacts keywords mapping to dictionaries containing inforation
        for each of them
        """

    def execute_with_snapshot(self) -> ModelSnapshot:
        """Method executing a model and returning a model snapshot

        :return: an initialized ModelSnapshot instance
        """
        outputs_dct = self.execute()
        return self._initialize_model_snapshot(outputs_dct)

    def _initialize_model_snapshot(self, outputs_dct: dict) -> ModelSnapshot:
        """Method initializing a model snapshot from model execution output dictionary and additional pieces of
        information.

        :param outputs_dct: a dictionary with figures of merit and artefacts
        :return: an initialized ModelSnapshot instance
        """
        artefacts = read_artefacts(
            outputs_dct, self.model_dependency_config.get_root_dir()
        )
        figures_of_merit = read_figures_of_merit(outputs_dct)
        return ModelSnapshot(
            name=self.model_dependency_config.get_model_name_without_ext(),
            abs_path=self.model_dependency_config.model_abs_path,
            execution_time=str(datetime.datetime.now()),
            modification_timestamp=int(
                os.path.getmtime(self.model_dependency_config.model_abs_path)
            ),
            input_parameters=self.input_parameters,
            artefacts=artefacts,
            figures_of_merit=figures_of_merit,
        )


def read_artefacts(outputs_dct: dict, model_dir: str) -> dict:
    """Function reading dictionaries and returning a dictionary. The dictionary is mapping from artefact name to its
    text content

    :param outputs_dct: a dictionary with figures of merit and artefacts
    :param model_dir: a root directory of a model
    :return: an empty dictionary if the artefact key is missing, otherwise a dictionary with artefact contents
    """
    if ARTEFACTS_KEYWORD not in outputs_dct:
        return {}

    return {
        artefact: read_artefact_as_text(
            model_dir, outputs_dct[ARTEFACTS_KEYWORD][artefact]
        )
        for artefact in outputs_dct[ARTEFACTS_KEYWORD]
    }


def read_figures_of_merit(outputs_dct: dict) -> dict:
    """Function reading a dictionary with figures of merit from the output dictionary

    :param outputs_dct: a dictionary with figures of merit and artefacts
    :return: an empty dictionary if the figure of merit key is missing, otherwise a dictionary with figures of merit
    """
    if FIGURES_OF_MERIT_KEYWORD not in outputs_dct:
        return {}

    return outputs_dct[FIGURES_OF_MERIT_KEYWORD]


def read_artefact_as_text(root_dir: str, artefact_rel_path: str) -> str:
    """Function reading an artefact file as text. The artefact path is constructed from the root directory and the
    relative artefact path.

    :param root_dir: root directory
    :param artefact_rel_path: relative artefact path w.r.t. the root directory
    :return: a string with artefact file content
    """
    artefact_abs_path = os.path.join(root_dir, artefact_rel_path)
    return text_file.read(artefact_abs_path)



from collections import defaultdict
from typing import Dict, Set

import pydantic
from pydantic import BaseModel
import yaml

from pymbse.query.config.hash_functions import hash_dict
from pymbse.query.config.model_type import ModelType
from pymbse.query.config.notebook_script_config import (
    NotebookScriptConfig,
    RootModelMetadataForHash,
)


class DatabaseConfig(BaseModel):
    """Class storing database configuration"""

    ip: str
    port: int

    def get_ip_with_port(self):
        return "http://%s:%d" % (self.ip, self.port)


class PyMBSEConfig(BaseModel):
    """Class putting together database configuration and model configurations"""

    database: DatabaseConfig
    models: Dict[str, NotebookScriptConfig]

    @pydantic.validator("models", pre=True)
    def validate(cls, models: dict):
        models_dct = {}
        for key, value in models.items():
            model_type = str(ModelType.NOTEBOOK_SCRIPT).split(".")[-1]
            if value["model_type"] == model_type:
                models_dct[key] = NotebookScriptConfig(**value)
        return models_dct

    def hash_root_model_metadata(self, root_model: str, input_parameters: dict, input_files: dict) -> str:
        return hash_dict(
            self.create_root_model_metadata(root_model, input_parameters, input_files).dict()
        )

    def create_root_model_metadata(
        self, root_model: str, input_parameters: dict, input_files: dict
    ) -> RootModelMetadataForHash:
        root_model_config = self.models[root_model]
        model_dependency_order_to_hash = (
            self._create_model_dependency_order_to_metadata_hash_dict(root_model)
        )
        return root_model_config.create_model_metadata().upcast_to_root(
            input_parameters, input_files, model_dependency_order_to_hash
        )

    def _create_model_dependency_order_to_metadata_hash_dict(self, root_model) -> dict:
        """Method creating a dictionary mapping from a model dependency order element to its hash

        :param root_model: name of a model for which a snapshot is hashed
        :return: a dictionary mapping from model dependency to its hash
        """
        model_dependency_dict = self._convert_model_dependency_tree_to_dict(root_model)
        model_dependency_order = convert_model_dependency_dict_to_order(
            model_dependency_dict
        )
        model_dependency_order_without_root_model = [
            mdo for mdo in model_dependency_order if mdo != root_model
        ]
        mdo_to_hash = {}
        for model_dependency in model_dependency_order_without_root_model:
            mdo_to_hash[model_dependency] = self.models[
                model_dependency
            ].hash_model_metadata()

        return mdo_to_hash

    def _convert_model_dependency_tree_to_dict(self, root_model: str) -> dict:
        """Method converting a model dependency tree given by the model configuration to a dictionary mapping tree level
        (0 being the tree root) to a list of model names at that level. The tree root is given by the model variable

        :param root_model: a model name for which the dependency tree is converted. It is the tree root.
        :return: a dictionary mapping tree level to list of models at that level
        """
        mdo_dct = defaultdict(lambda: [])

        def mdo(model: str, level: int) -> None:
            """Function converting a dependency tree into a model dependency order (mdo). The function recursively
            traverses the tree and builds a dictionary mapping tree level (0-based index, where 0 is the root) to a list
            of models at that level. Models at a given level are executed independently.

            :param model: model name for which dependencies are expanded
            :param level: tree level (0-based index, where 0 is the root)
            """
            mdo_dct[level].append(model)

            needs = self.models[model].needs
            for need in needs:
                mdo(need, level + 1)

        mdo(root_model, 0)
        return mdo_dct


def convert_model_dependency_dict_to_order(model_dependency_dict: dict) -> Set[str]:
    """Function converting a model dependency dictionary, mapping from tree level (from root to leaves) to a list of
    models at that level, to a model dependency order (mdo) represented as a set. The mdo goes from the bottom-most
    models to the root

    :param model_dependency_dict: model dependency dictionary, mapping from tree level (from root to leaves) to a list
    of models at that level
    :return: a set of models representing an mdo
    """
    model_dependency_order = set()
    for level in sorted(model_dependency_dict.keys(), reverse=True):
        model_dependency_order |= set(model_dependency_dict[level])

    return model_dependency_order


def load_pymbse_config(pymbse_config_path: str) -> PyMBSEConfig:
    """Function loading PyMBSE configuration from a YAML file

    :param pymbse_config_path: an absolute path to a query configuration
    :return: an initialized PyMBSEConfig instance
    """
    with open(pymbse_config_path) as f:
        return PyMBSEConfig(**yaml.safe_load(f))



import hashlib
import json
import os
import warnings
from pathlib import Path
from typing import Union

from pymbse.query.config.numpy_encoder import NumpyEncoder  # type: ignore


def hash_dict(dict_to_hash: dict) -> str:
    """Function hashing a dictionary with md5 algorithm

    :param dict_to_hash: a dictionary with content to be hashed
    :return: an md5 hash of a dictionary
    """
    return hashlib.md5(
        json.dumps(dict_to_hash, sort_keys=True, cls=NumpyEncoder).encode("utf-8")
    ).hexdigest()


def hash_file(abs_path: Union[Path, str]) -> str:
    """Function hashing a file based on its content with md5 hash algorithm
    :param abs_path: an absolute path to a file
    :return: a string with md5 hash
    """
    if not os.path.isfile(abs_path):
        warnings.warn(
            f"Can't hash an input file {abs_path} as it does not exist. "
            f"Either create the file or change model_configuration.yml definition."
        )
        return hashlib.md5("".encode("utf-8")).hexdigest()

    with open(abs_path, "rb") as file:
        return hashlib.md5(file.read()).hexdigest()



from pymbse.query.config.model_type import ModelType
from pymbse.query.model_api.notebook_script_api import (
    NotebookModelAPI,
    ScriptModelAPI,
    NotebookScriptModelAPI,
)

model_type_to_api = {
    ModelType.NOTEBOOK: NotebookModelAPI,
    ModelType.SCRIPT: ScriptModelAPI,
    ModelType.NOTEBOOK_SCRIPT: NotebookScriptModelAPI,
}



import json
from typing import Any, Union

import numpy as np  # type: ignore


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Union[int, float, list, Any]:
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return json.JSONEncoder.default(self, obj)



import os
import warnings
from typing import List
from urllib.parse import urljoin

import requests


def init_upload_and_run(base_url, model_name: str, input_files: List[str]) -> tuple:

    # init model
    timestamp = init_model(base_url, model_name)

    # upload unput files
    upload_input_files(base_url, model_name, timestamp, input_files)

    # run model
    return run_model(base_url, model_name, timestamp)


def init_model(base_url: str, model_name: str) -> str:
    url_init = urljoin(base_url, f"model/{model_name}")
    response_init = requests.post(url_init)
    if response_init.status_code != 200:
        raise RuntimeError(response_init.json()["detail"]["output"])

    return response_init.json()["timestamp"]


def upload_input_files(base_url, model_name, timestamp, input_files) -> None:
    url_upload = urljoin(base_url, "model")
    files = [("files", open(input_file)) for input_file in input_files]
    data = {"model_name": model_name, "timestamp": timestamp}

    response = requests.post(url_upload, files=files, data=data)

    if response.status_code != 200:
        raise RuntimeError(response.json()["detail"]["output"])


def run_model(base_url, model_name, timestamp) -> tuple:
    url = urljoin(base_url, f"model/{model_name}/{timestamp}/run")
    response = requests.post(url)

    if response.status_code != 200:
        raise RuntimeError(response.json()["detail"]["output"])

    output_lines = response.json()["output"].split("\n")
    model_name = response.json()["model_name"]
    timestamp = response.json()["timestamp"]
    return model_name, timestamp, output_lines


def download_artefact(base_url, model_name, timestamp, input_dir, artefact):
    artefact_url = urljoin(base_url, f"artefact/{model_name}/{timestamp}/{artefact}")
    response_artefact = requests.get(artefact_url)
    if response_artefact.status_code != 200:
        warnings.warn("The requested artefact %s is not available!" % artefact)
    
    with open(os.path.join(input_dir, artefact), "wb") as file:
        file.write(response_artefact.content)



from pymbse.query.config.pymbse_config import load_pymbse_config
from pymbse.query.model_api.model_api import ModelAPI
from pymbse.query.model_api.model_api_factory import model_type_to_api


class PyMBSE:
    """A class providing a model query mechanism. The models are embedded into Jupyter notebooks and expose the
    following endpoints:
    - get_html_report()
    - get_ipynb_report()
    - get_figures_of_merit()
    For notebook models
    - get_artefacts()
    - get_artefact(artefact_name)

    For the query mechanism to work, a notebook should provide an input and output functionality.
    - Input:
      - a notebook cell with input parameters is marked with a tag "parameters"
      - the input parameter cell contains only variable assignments, e.g., a = 3. Thus, the input parameter cell should
      not contain comments, iPython magic, import statements, etc.

    - Output
      - a notebook cell with outputs does not require any tag
      - the outputs are returned with a statement `sb.glue('model_results', data=fom_dct, encoder='json')`, where
      fom_dct is a dictionary composed of two keys: `fom` and `artefacts`:
        - the value of the `fom` key  contains a figure of merit key-value dictionary
        - the value of the `artefacts` key contains an artefact key-value dictionary with keys being artefact file
        names and values corresponding to relative file paths w.r.t. the root model directory

    """

    def __init__(
        self,
        config_path: str,
        source_model: str,
        target_model: str,
        input_parameters: dict = None,
        input_files: dict = None
    ) -> None:
        """

        :param config_path: a path to the model dependency config
        :param source_model: a source model from which a PyMBSE query is made (used to check whether the target model is
        in its the dependency). If the source_model is an empty string, then the target_model is called as a standalone
        model. In this case, there is no check of the dependency.
        :param target_model: name of a model to be executed.
        :param input_parameters: a dictionary with input parameters for the target_model
        """
        self.configuration = load_pymbse_config(config_path)
        self.source_model = source_model
        self.target_model = target_model

        is_source_model_in_models = self.source_model not in self.configuration.models
        if self.source_model != "" and is_source_model_in_models:
            raise KeyError(
                "Source model %s not present in model configuration!"
                % self.source_model
            )

        if (
            self.source_model != ""
            and self.target_model
            not in self.configuration.models[self.source_model].needs
        ):
            raise KeyError(
                "Target model %s not present in source model dependencies!"
                % self.target_model
            )

        self.input_parameters = {} if input_parameters is None else input_parameters
        self.input_files = {} if input_files is None else input_files

    def build(self) -> ModelAPI:
        """Method building a ModelAPI instance based on model configuration. It passes pieces of information needed for
        model execution, namely:
        - target model config
        - target model snapshot hash (for caching)
        - input parameters
        - input files

        :return: a ModelAPI implementation matching the model type
        """
        target_model_config = self.configuration.models[self.target_model]
        ip_with_port = self.configuration.database.get_ip_with_port()
        # create hash for a model
        root_model_metadata_hash = self.configuration.hash_root_model_metadata(
            root_model=self.target_model, input_parameters=self.input_parameters, input_files=self.input_files
        )
        model_api_class = model_type_to_api[target_model_config.model_type]
        return model_api_class(
            root_model_config=target_model_config,
            ip_with_port=ip_with_port,
            root_model_metadata_hash=root_model_metadata_hash,
            input_parameters=self.input_parameters,
            input_files=self.input_files
        )



import os
import sys
from typing import Callable

import papermill as pm
import scrapbook as sb
from pymbse.commons import text_file

from pymbse.query.model_api.executor.executor import Executor

MODEL_RESULTS_KEYWORD = "model_results"


class ScriptExecutor(Executor):
    """ScriptExecutor class is implementing Executor abstract class and, particularly, the execute method for
    scripts obtained from notebooks.

    """

    def execute(self) -> dict:
        write_input_files(self.model_dependency_config.get_root_dir(), self.input_files)

        script = self.model_dependency_config.get_script_name_without_ext()
        cwd = os.getcwd()

        os.chdir(self.model_dependency_config.get_root_dir())

        # Unload a module to allow for its modification between subsequent calls
        if script in sys.modules.keys():
            sys.modules.pop(script)
        run = ScriptExecutor._import_function_from_module(script, "run_" + script)
        try:
            fom_model = run(**self.input_parameters)
        except:
            raise
        finally:
            os.chdir(cwd)

        return fom_model

    @staticmethod
    def _import_function_from_module(module: str, function: str) -> Callable:
        return getattr(__import__(module), function)


class NotebookExecutor(Executor):
    """NotebookExecutor class is implementing Executor abstract class and, particularly, the execute method for
    notebooks.

    """

    def execute(self) -> dict:
        write_input_files(self.model_dependency_config.get_root_dir(), self.input_files)
        model_out_abs_path = self.model_dependency_config.get_ipynb_out_abs_path()
        pm.execute_notebook(
            self.model_dependency_config.model_abs_path,
            model_out_abs_path,
            cwd=self.model_dependency_config.get_root_dir(),
            parameters=self.input_parameters,
        )

        return sb.read_notebook(model_out_abs_path).scraps[MODEL_RESULTS_KEYWORD].data


def write_input_files(root_dir: str, input_files: dict) -> None:
    for input_file_rel_path, input_file_content in input_files:
        input_file_abs_path = os.path.join(root_dir, input_file_rel_path)
        text_file.write(input_file_abs_path, input_file_content)


import os

import nbformat
from nbconvert import HTMLExporter

import pymbse.commons.json_file as json_file
import pymbse.commons.text_file as text_file

IMPORT_KEYWORD = "import"
FROM_KEYWORD = "from"
PARAMETERS_KEYWORD = "parameters"
TAGS_KEYWORD = "tags"
METADATA_CELL_KEYWORD = "metadata"
SB_GLUE_OUTPUT_KEYWORD = "sb.glue"
SOURCE_CELL_KEYWORD = "source"
CODE_KEYWORD = "code"
CELLS_KEYWORD = "cells"
CELL_TYPE_KEYWORD = "cell_type"


def convert_ipynb_to_html(ipynb_abs_path: str, html_abs_path: str) -> None:
    """Function converting an ipynb file executed with papermill (containing the outputs) to an html file

    :param ipynb_abs_path: an absolute path to an ipynb output notebook
    :param html_abs_path: an absolute path to an html report
    """
    with open(ipynb_abs_path) as f:
        nb = nbformat.read(f, as_version=4)

    config = {
        "TemplateExporter": {"exclude_input": True},
        "TagRemovePreprocessor": {"remove_all_outputs_tags": ["skip_output"]},
    }
    html_exporter = HTMLExporter(config)
    body, _ = html_exporter.from_notebook_node(nb)
    decoded_body = body.encode("ascii", "ignore").decode()
    with open(html_abs_path, "w") as f:
        f.write(decoded_body)


def create_script_if_missing_or_outdated(
    notebook_abs_path: str, script_abs_path: str, model_name_without_ext: str
) -> None:
    """Function creating a script from a notebook if the script is either missing or outdated (notebook was updated
    after the script creation).

    :param notebook_abs_path: an absolute path to the model notebook
    :param script_abs_path: an absolute path to the script
    :param model_name_without_ext: name of a model without an extension
    """
    if is_script_missing_or_outdated(notebook_abs_path, script_abs_path):
        convert_notebook_to_script(
            notebook_abs_path, model_name_without_ext, script_abs_path
        )


def is_script_missing_or_outdated(notebook_abs_path: str, script_abs_path: str) -> bool:
    """Function checking whether a script is either missing or outdated

    :param notebook_abs_path: an absolute path to a model notebook
    :param script_abs_path: an absolute path to a script created from model notebook
    :return: True if script is either missing or outdated, otherwise True
    """
    if is_script_missing(script_abs_path):
        return True
    elif is_script_outdated(notebook_abs_path, script_abs_path):
        os.remove(script_abs_path)
        return True
    else:
        return False


def is_script_missing(script_abs_path: str) -> bool:
    """Function checking whether a script is missing

    :param script_abs_path: an absolute path to a script created from model notebook
    :return: True if a script is missing, otherwise False
    """
    return not os.path.isfile(script_abs_path)


def is_script_outdated(notebook_abs_path: str, script_abs_path: str) -> bool:
    """Function checking whether a script is outdated w.r.t. a notebook.
    This happens if the model last modification time is greater or equal to the one of a notebook.
    The equality of the modification time may occur if an outdated script and the model were moved at the same time.
    e.g., while cloning a repository containing mis-synchronized model and script, or copying the two.
    It is assumed that the copying should happen in less then a second. Thus the comparison is carried out on integer
    modification times.

    :param notebook_abs_path: an absolute path to a model notebook
    :param script_abs_path: an absolute path to a script created from model notebook
    :return: True if script is outdated, otherwise False
    """
    model_last_modification_time = os.path.getmtime(notebook_abs_path)
    script_last_modification_time = os.path.getmtime(script_abs_path)

    return int(model_last_modification_time) >= int(script_last_modification_time)


def convert_notebook_to_script(
    notebook_abs_path: str, notebook_name: str, script_abs_path: str
) -> None:
    """Function converting a notebook into a script equivalent. The notebook is executed as a function.
    The function name is given by as run_<notebook_name>. The input arguments are taken from cells with `parameters`
    tags. The return value is taken from the sb.glue code.

    :param notebook_abs_path: an absolute path to a model notebook
    :param notebook_name: name of a notebook without extension
    :param script_abs_path: an absolute path to a script created from model notebook
    """
    # Parse notebook
    notebook = json_file.read(notebook_abs_path)
    cells_with_parameters, cells_with_code = separate_parameters_and_code_cells(
        notebook
    )
    parameter_lines = extract_parameter_lines(cells_with_parameters)
    code_lines = extract_code_lines(cells_with_code)
    output_variable = extract_output_variable(cells_with_code)
    cleaned_code_lines = remove_ipython_magic_and_sys_path_apppend(code_lines)

    # Construct script
    function_parameters = ", ".join(parameter_lines)
    output_lines = []
    output_lines.append("\n")
    output_lines.append("def run_%s_script(%s):" % (notebook_name, function_parameters))

    for cleaned_code_line in cleaned_code_lines:
        output_lines.append("\t" + cleaned_code_line)

    output_lines.append("\t" + "return " + output_variable)

    text_file.writelines(script_abs_path, output_lines)


def separate_parameters_and_code_cells(notebook: dict) -> tuple:
    """Function separating cells with parameters from those with code

    :param notebook: a dictionary witho notebook contents
    :return: a two-element tuple containing a list of cells with parameters and list of cells with code
    """
    cells_with_all_code = [
        cell
        for cell in notebook[CELLS_KEYWORD]
        if cell[CELL_TYPE_KEYWORD] == CODE_KEYWORD
    ]
    cells_with_parameters = []
    cells_with_code = []

    for cell in cells_with_all_code:
        if TAGS_KEYWORD in cell[METADATA_CELL_KEYWORD] and (
            PARAMETERS_KEYWORD in cell[METADATA_CELL_KEYWORD][TAGS_KEYWORD]
        ):
            cells_with_parameters.append(cell)
        else:
            cells_with_code.append(cell)

    return cells_with_parameters, cells_with_code


def extract_parameter_lines(cells_with_parameters: list) -> list:
    """Function extracting parameter lines from a list of cells with parameters. In case there is a comment in a
    parameter line, an AttributeError is raised.

    :param cells_with_parameters: a list of cells with parameters
    :return: a list of lines with parameters
    """
    parameters = []
    for cell in cells_with_parameters:
        for line in cell[SOURCE_CELL_KEYWORD]:
            if "#" in line:
                raise AttributeError(
                    "No commented lines are allowed in the Input Parameters cell!"
                )
            parameters.append(line.replace("\n", ""))

    return parameters


def extract_code_lines(cells_with_code: list) -> list:
    """Function extracting code lines from lines with code

    :param cells_with_code: a list of cells with code
    :return: a list of code lines
    """
    code_lines = []
    for cell in cells_with_code:
        for line in cell[SOURCE_CELL_KEYWORD]:
            if SB_GLUE_OUTPUT_KEYWORD not in line:
                code_lines.append(line.replace("\n", ""))
    return code_lines


def extract_output_variable(cells_with_code: list) -> str:
    """Function extracting an output variable name from cells with code

    :param cells_with_code: a list of cells with code
    :return: an output variable name
    """
    output_line = ""
    for cell in cells_with_code:
        for line in cell[SOURCE_CELL_KEYWORD]:
            if SB_GLUE_OUTPUT_KEYWORD in line:
                output_line = line

    output_variable = output_line.split(",")[1]
    if "=" in output_variable:
        output_variable = output_variable.split("=")[-1]

    return output_variable


def remove_ipython_magic_and_sys_path_apppend(code_lines: list) -> list:
    """Function separating code lines into those with import statements and those with remaining code

    :param code_lines: a list of code lines for separation
    :return: a two-element tuple containing a list of import lines and a list of remaining code lines
    """
    cleaned_code_lines = []

    for code_line in code_lines:
        if not (code_line.startswith("%") or ("sys.path.append" in code_line)):
            cleaned_code_lines.append(code_line)

    return cleaned_code_lines



import os
from pathlib import Path
from typing import Set, List

import pydantic
from pydantic import BaseModel

from pymbse.query.config.model_type import ModelType
from pymbse.query.config.hash_functions import hash_file, hash_dict


class ModelMetadataForHash(BaseModel):
    """Class providing a container to store notebook metadata"""

    model_file_hash: str
    input_file_to_hash: dict

    def upcast_to_root(
        self, input_parameters: dict, input_files: dict, model_dependency_order_to_hash: dict
    ) -> "RootModelMetadataForHash":
        return RootModelMetadataForHash(
            **self.dict(),
            input_parameters=input_parameters,
            input_files=input_files,
            model_dependency_order_to_hash=model_dependency_order_to_hash
        )


class RootModelMetadataForHash(ModelMetadataForHash):
    """Class providing a container to store model notebook metadata. It enhances the NotebookMetadata class with a
    dictionary of input parameters and a dictionary with model dependency hashes

    """

    input_parameters: dict
    input_files: dict
    model_dependency_order_to_hash: dict


class ScriptConfig(BaseModel):
    """Class storing script model config"""

    model_abs_path: str
    model_type: ModelType
    input_files: Set[str]
    needs: List[str]

    class Config:
        use_enum_values = True

    @pydantic.validator("model_type", pre=True)
    def validate_enum_field(cls, model_type: str):
        return ModelType(model_type)

    def get_root_dir(self) -> str:
        """Method returning the root directory of a model

            >>> from pymbse.query.config.notebook_script_config import ScriptConfig
            >>> config = ScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.py', model_type=ModelType.SCRIPT,
            >>> input_files=set(), needs=[])
            >>> config.get_root_dir()
            "C:\\magnetic"


        :return: string with the root directory of a model
        """
        return os.path.dirname(self.model_abs_path)

    def get_model_name_without_ext(self) -> str:
        """Method returning the model name without extension

            >>> from pymbse.query.config.notebook_script_config import ScriptConfig
            >>> config = ScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.py', model_type=ModelType.SCRIPT,
            >>> input_files=set(), needs=[])
            >>> config.get_model_name_without_ext()
            "ROXIE"

        :return: string with model name without extension
        """
        return Path(self.model_abs_path).stem

    def get_script_name(self) -> str:
        """Method returning the script name for a notebook

            >>> from pymbse.query.config.notebook_script_config import ScriptConfig
            >>> config = ScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.py', model_type=ModelType.SCRIPT,
            >>> input_files=set(), needs=[])
            >>> config.get_script_name()
            "ROXIE.py"

        :return: string with script name
        """
        _, path_tail = os.path.split(self.model_abs_path)
        return path_tail

    def get_script_name_without_ext(self) -> str:
        """Method returning the script name for a notebook without extension

            >>> from pymbse.query.config.notebook_script_config import ScriptConfig
            >>> config = ScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.py',
            >>> model_type=ModelType.NOTEBOOK_SCRIPT, input_files=set(), needs=[])
            >>> config.get_script_name_without_ext()
            "ROXIE"

        :return: string with script name
        """
        return self.get_model_name_without_ext()

    def get_script_abs_path(self) -> str:
        """Method returning the absolute path of a script name for a notebook

            >>> from pymbse.query.config.notebook_script_config import ScriptConfig
            >>> from pymbse.query.config.model_type import ModelType
            >>> config = ScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.py',
            >>> model_type=ModelType.SCRIPT, input_files=set(), needs=[])
            >>> config.get_script_abs_path()
            "C:\\magnetic\\ROXIE.py"

        :return: string with absolute path of script
        """
        return self.model_abs_path

    def create_input_file_to_hash_dict(self) -> dict:
        """Method creating a dictionary with hashed inputs. The dictionary is mapping from input name to content hash

        :return: a dictionary mapping from input file name to its content hash
        """
        current_model_root_dir = self.get_root_dir()
        input_file_to_hash = {}
        for input_file in self.input_files:
            input_file_path = (
                Path(current_model_root_dir) / Path(input_file)
            ).resolve()
            input_file_to_hash[input_file] = hash_file(input_file_path)

        return input_file_to_hash

    def hash_model_metadata(self) -> str:
        return hash_dict(self.create_model_metadata().dict())

    def create_model_metadata(self) -> ModelMetadataForHash:
        input_file_to_hash = self.create_input_file_to_hash_dict()
        return ModelMetadataForHash(
            model_file_hash=hash_file(self.model_abs_path),
            input_file_to_hash=input_file_to_hash,
        )


class NotebookScriptConfig(ScriptConfig):
    """Class storing notebook/script model config"""

    def get_script_name(self) -> str:
        """Method returning the script name for a notebook

            >>> from pymbse.query.config.notebook_script_config import NotebookScriptConfig
            >>> from pymbse.query.config.model_type import ModelType
            >>> config = NotebookScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.ipynb',
            >>> model_type=ModelType.NOTEBOOK_SCRIPT, input_files=set(), needs=[])
            >>> config.get_script_name()
            "ROXIE_script.py"

        :return: string with script name
        """
        script_name_without_ext = self.get_script_name_without_ext()
        return "%s.py" % script_name_without_ext

    def get_script_name_without_ext(self) -> str:
        """Method returning the script name for a notebook without extension

            >>> from pymbse.query.config.notebook_script_config import NotebookScriptConfig
            >>> config = NotebookScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.ipynb',
            >>> model_type=ModelType.NOTEBOOK_SCRIPT, input_files=set(), needs=[])
            >>> config.get_script_name_without_ext()
            "ROXIE_script"

        :return: string with script name
        """
        model_name = self.get_model_name_without_ext()
        return "%s_script" % model_name

    def get_script_abs_path(self) -> str:
        """Method returning the absolute path of a script name for a notebook

            >>> from pymbse.query.config.notebook_script_config import NotebookScriptConfig
            >>> config = NotebookScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.ipynb',
            >>> model_type=ModelType.NOTEBOOK_SCRIPT, input_files=set(), needs=[])
            >>> config.get_script_abs_path()
            "C:\\magnetic\\ROXIE_script.py"

        :return: string with absolute path of script
        """
        root_dir = self.get_root_dir()
        script_name = self.get_script_name()
        return os.path.join(root_dir, script_name).replace("\\", "/")

    def get_html_out_abs_path(self) -> str:
        """Method returning the absolute path of an HTML report

            >>> from pymbse.query.config.notebook_script_config import NotebookScriptConfig
            >>> config = NotebookScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.ipynb',
            >>> model_type=ModelType.NOTEBOOK_SCRIPT, input_files=set(), needs=[])
            >>> config.get_html_out_abs_path()
            "C:\\magnetic\\ROXIE_out.html"

        :return: string with absolute path of HTML report
        """
        root_dir = self.get_root_dir()
        model_name = self.get_model_name_without_ext()
        html_name = "%s_out.html" % model_name
        return os.path.join(root_dir, html_name).replace("\\", "/")

    def get_ipynb_out_abs_path(self) -> str:
        """Method returning the absolute path of an ipynb report

            >>> from pymbse.query.config.notebook_script_config import NotebookScriptConfig
            >>> config = NotebookScriptConfig(model_abs_path='C:\\magnetic\\ROXIE.ipynb',
            >>> model_type=ModelType.NOTEBOOK_SCRIPT, input_files=set(), needs=[])
            >>> config.get_ipynb_out_abs_path()
            "C:\\magnetic\\ROXIE_out.ipynb"

        :return: string with absolute path of ipynb report
        """
        root_dir = self.get_root_dir()
        model_name = self.get_model_name_without_ext()
        ipynb_name = "%s_out.ipynb" % model_name
        return os.path.join(root_dir, ipynb_name).replace("\\", "/")



from abc import ABC, abstractmethod

from pydantic import BaseModel

from pymbse.query.config.pymbse_config import NotebookScriptConfig
from pymbse.query.model_api.snapshot.model_snapshot import ModelSnapshot


class ModelAPI(BaseModel, ABC):
    root_model_config: NotebookScriptConfig
    ip_with_port: str
    root_model_metadata_hash: str
    input_parameters: dict
    input_files: dict

    def get_figures_of_merit(self) -> dict:
        return self.execute().figures_of_merit

    def get_artefacts(self) -> dict:
        return self.execute().artefacts

    def get_artefact(self, artefact_name: str) -> str:
        model_snapshot = self.execute()
        if artefact_name not in model_snapshot.artefacts:
            print(
                "An artefact %s is missing. Returning an empty string." % artefact_name
            )
            return ""

        return model_snapshot.artefacts[artefact_name]

    @abstractmethod
    def execute(self) -> ModelSnapshot:
        """Abstract method executing a model to be implemented by subclasses."""



import json
from typing import Callable, Union

import requests
from pymbse.commons import text_file

from pymbse.query.config.numpy_encoder import NumpyEncoder
from pymbse.query.model_api.executor.notebook_script_executor import (
    ScriptExecutor,
    NotebookExecutor,
)
from pymbse.query.model_api.executor.notebook_converter import (
    create_script_if_missing_or_outdated,
)

from pymbse.query.model_api.executor.notebook_converter import convert_ipynb_to_html
from pymbse.query.model_api.model_api import ModelAPI
from pymbse.query.model_api.snapshot.model_snapshot import (
    ModelSnapshot,
    HashedModelSnapshot,
)

API_ENDPOINT = "api/model_snapshots"


def post_model_snapshot_to_cache_decorator(execute_notebook: Callable) -> Callable:
    """Decorator posting a model snapshot to model_api with notebook execution. The notebook itself is not uploaded
    due to its, potentially, excessive size and MongoDb message limit of 16 Mb.

    :param execute_notebook: a function executing a notebook and returning model snapshot
    :return: a cached execute_script function that either stores the model snapshot or not
    """

    def wrapper(model_api: ModelAPI) -> None:
        """Wrapper function executing a notebook

        :param model_api: ModelAPI instance
        """
        cached_model_snapshot = get_cached_model_snapshot(
            model_api.ip_with_port, model_api.root_model_metadata_hash
        )
        if cached_model_snapshot is None:
            executed_model_snapshot = execute_notebook(model_api)
            post_model_snapshot_to_cache(
                model_api.ip_with_port,
                executed_model_snapshot,
                model_api.root_model_metadata_hash,
            )
        else:
            execute_notebook(model_api)

    return wrapper


def get_model_snapshot_from_cache_decorator(execute_script: Callable) -> Callable:
    """Decorator acting on a ModelAPI method and providing a caching mechanism. If a model snapshot model_api is present in a
    database, then its snapshot is returned, otherwise the script is executed and the snapshot stored in the database.

    :param execute_script: an execute_script function
    :return: a cached execute_script function that either executes the function or returns a cached model snapshot
    """

    def wrapper(model_api: ModelAPI) -> ModelSnapshot:
        """Function wrapping the model_api functionality

        :param model_api: ModelAPI instance
        :return: a ModelSnapshot instance either obtained from the model_api or from script execution
        """

        cached_model = get_cached_model_snapshot(
            model_api.ip_with_port, model_api.root_model_metadata_hash
        )
        if cached_model is None:
            print(
                "Executing %s as its model snapshot is not yet cached."
                % model_api.root_model_config.model_abs_path
            )
            cached_model = execute_script(model_api)
            post_model_snapshot_to_cache(
                model_api.ip_with_port, cached_model, model_api.root_model_metadata_hash
            )

        return cached_model

    return wrapper


def get_cached_model_snapshot(
    ip_with_port: str, model_snapshot_hash: str
) -> Union[None, ModelSnapshot]:
    """Function getting a cached model from a database

    :param ip_with_port: a path to the database with an ip and port, e.g., localhost:8000
    :param model_snapshot_hash: a model_api of the model snapshot, it is the primary key for the model_api mechanism
    :return: None if a model_api is not present in the database, otherwise a ModelSnapshot instance
    """
    url = "%s/%s/%s" % (ip_with_port, API_ENDPOINT, model_snapshot_hash)
    response = requests.request("GET", url)
    if response.json() is None or response.status_code != 200:
        return None
    else:
        return HashedModelSnapshot(**response.json()).downcast_to_model_snapshot()


def post_model_snapshot_to_cache(
    ip_with_port: str, model_snapshot: ModelSnapshot, model_snapshot_hash: str
) -> None:
    """Function posting a model snapshot with model_api to the model_api database

    :param ip_with_port: a path to the database with an ip and port, e.g., localhost:8000
    :param model_snapshot: a model snapshot to be cached
    :param model_snapshot_hash: a model_api of the model snapshot, it is the primary key for the model_api mechanism
    """
    url = "%s/%s/" % (ip_with_port, API_ENDPOINT)
    payload_dct = model_snapshot.upcast_to_hashed_model_snapshot(
        model_snapshot_hash
    ).dict()
    payload_json = json.loads(json.dumps(payload_dct, cls=NumpyEncoder))
    response = requests.request("POST", url, json=payload_json)
    if response.status_code != 201:
        raise ConnectionError("Could not post model snapshot to url: %s" % url)


class NotebookModelAPI(ModelAPI):
    def get_html_report(self) -> str:
        """Method returning an HTML report content. It executes a notebook with papermill and returns an HTML file
        content.

        :return: a string containing the HTML report content
        """
        self.execute_without_cache()
        convert_ipynb_to_html(
            ipynb_abs_path=self.root_model_config.get_ipynb_out_abs_path(),
            html_abs_path=self.root_model_config.get_html_out_abs_path(),
        )
        return text_file.read(self.root_model_config.get_html_out_abs_path())

    def get_ipynb_report(self) -> str:
        """Method returning an ipynb report content. It executes a notebook with papermill and returns an ipynb file
        content.

        :return: a string containing the ipynb report content
        """
        self.execute_without_cache()
        return text_file.read(self.root_model_config.get_ipynb_out_abs_path())

    @post_model_snapshot_to_cache_decorator
    def execute_without_cache(self) -> ModelSnapshot:
        return NotebookExecutor(
            model_dependency_config=self.root_model_config,
            input_parameters=self.input_parameters,
            input_files=self.input_files
        ).execute_with_snapshot()

    @get_model_snapshot_from_cache_decorator
    def execute(self) -> ModelSnapshot:
        return NotebookExecutor(
            model_dependency_config=self.root_model_config,
            input_parameters=self.input_parameters,
            input_files=self.input_files
        ).execute_with_snapshot()


class ScriptModelAPI(ModelAPI):
    @get_model_snapshot_from_cache_decorator
    def execute(self) -> ModelSnapshot:
        return ScriptExecutor(
            model_dependency_config=self.root_model_config,
            input_parameters=self.input_parameters,
            input_files=self.input_files
        ).execute_with_snapshot()


def create_script_if_missing_or_outdated_decorator(
    script_dependent_function: Callable,
) -> Callable:
    """

    :param script_dependent_function: a function executing
    :return: a script_dependent_function with ensured script presence
    """

    def wrapper(self: ModelAPI, *args, **kwargs) -> Callable:
        """Wrapper function executing the script_dependent_function

        :param self: PyMBSE instance
        :param args: a list of arguments
        :param kwargs: a dictionary with keyword arguments
        :return: script_dependent_function call
        """
        notebook_abs_path = self.root_model_config.model_abs_path
        script_abs_path = self.root_model_config.get_script_abs_path()
        model_name_without_ext = self.root_model_config.get_model_name_without_ext()
        create_script_if_missing_or_outdated(
            notebook_abs_path=notebook_abs_path,
            script_abs_path=script_abs_path,
            model_name_without_ext=model_name_without_ext,
        )
        return script_dependent_function(self, *args, **kwargs)

    return wrapper


class NotebookScriptModelAPI(ScriptModelAPI, NotebookModelAPI):
    @create_script_if_missing_or_outdated_decorator
    def get_figures_of_merit(self) -> dict:
        return super().get_figures_of_merit()

    @create_script_if_missing_or_outdated_decorator
    def get_artefacts(self) -> dict:
        return super().get_artefacts()

    @create_script_if_missing_or_outdated_decorator
    def get_artefact(self, artefact_name: str) -> str:
        return super().get_artefact(artefact_name)



from copy import deepcopy

from pydantic import BaseModel


class ModelSnapshot(BaseModel):
    """Class providing a container for a model snapshot created after an execution of a model.
    It provides a function to upcast itself to a hashed model snapshot used for caching purposes.
    """

    name: str
    abs_path: str
    execution_time: str
    modification_timestamp: int
    input_parameters: dict
    artefacts: dict
    figures_of_merit: dict

    def upcast_to_hashed_model_snapshot(
        self, model_snapshot_hash: str
    ) -> "HashedModelSnapshot":
        """Method upcasting a model snapshot to its subclass with a model snapshot model_api

        :param model_snapshot_hash: a string with model snapshot model_api used for caching
        :return: an instance of HashedModelSnapshot
        """
        return HashedModelSnapshot(**self.__dict__, hash=model_snapshot_hash)


class HashedModelSnapshot(ModelSnapshot):
    """Class providing a container for hashed model snapshot used for communication with the model_api database.
    It provides a function to downcast a hashed model snapshot ot its superclass.

    """

    hash: str

    def downcast_to_model_snapshot(self) -> ModelSnapshot:
        """Method downcasting a HashedModelSnapshot instance into ModelSnapshot superclass to be used by PyMBSE

        :return: a ModelSnapshot instance
        """
        model_snapshot_dct = deepcopy(self.__dict__)
        del model_snapshot_dct["hash"]
        return ModelSnapshot(**model_snapshot_dct)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymbse',
 'pymbse.microservice',
 'pymbse.query',
 'pymbse.query.config',
 'pymbse.query.model_api',
 'pymbse.query.model_api.executor',
 'pymbse.query.model_api.snapshot']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'nbconvert>=6.5.0,<7.0.0',
 'papermill>=2.3.4,<3.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'pymbse-commons>=0.0.6,<0.0.7',
 'scrapbook>=0.5.0,<0.6.0']

setup_kwargs = {
    'name': 'pymbse',
    'version': '0.0.16',
    'description': '',
    'long_description': 'None',
    'author': 'mmaciejewski',
    'author_email': 'michal.maciejewski@ief.ee.ethz.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
